

import sys
import os
import re
import time
import math
import numpy as np
import multiprocessing as mp
import json

from modelpruner.models.fba_model import FbaModel
from modelpruner.core.protected_parts import ProtectedParts

global _fba_model
global _pfs


def _init_worker(fba_model, pfs):
    """Multiprocessing initializer for workers.

    Configure global parameters for the worker process

    :param fba_model: Fba model to apply FVA
    :type fba_model: Class FbaModel
    :param pfs: Protected Functions
    :type pfs: dict, key: function id, value: class ProtectedFunction
    """
    global _fba_model
    global _pfs

    _fba_model = fba_model.copy()
    _pfs = pfs


def _worker(rids):
    """Mulitprocessing worker to run FVA for selected reactions over all conditions.

    Checked that best performance achieved when worker performs FVA for
    all protected functions and a larger range for reactions.

    :param rids: reactions for FVA
    :type rids: list-like of str
    """
    global _fba_model
    global _pfs

    flux_min_pfs = np.zeros((len(_pfs), len(rids)))
    flux_max_pfs = np.zeros((len(_pfs), len(rids)))

    for idx, pf in enumerate(_pfs.values()):
        res = _fba_model.fva_pf_flux_ranges(pf, rids)
        if res['success'] is True:
            flux_min_pfs[idx] = res['min']
            flux_max_pfs[idx] = res['max']
        else:
            print(f'FBA for condition {pf.obj_id} is not feasible')
            flux_min_pfs[idx] = np.nan
            flux_max_pfs[idx] = np.nan
            break
    # delete linear problem
    # del _fba_model.delete_fba_lp()
    res = {'success': True, 'rids': rids, 'min_pfs': flux_min_pfs, 'max_pfs': flux_max_pfs}
    return res


class ModelPruner:

    def __init__(self, sbml_fname, protected_parts, resume=False, processes=100):
        """Init

        :param sbml_fname: pathname of original model (ending with '.xml')
        :type sbml_fname: str
        :param protected_parts: pathname of protected parts file or dict
        :type protected_parts: str or dict with protected parts information
        :param resume: Indicator if to resume pruning from stored snapshot
        :type resume: bool, optional (default: False)
        :param processes: maximal number of processes for multiprocessing
        :type processes: int (default: 100)
        """
        self.tolerance = 1e-7   # later accessible by set_params() ?
        self.cpus = min(processes, os.cpu_count())
        self.min_mp_total_flux = 1000
        self.full_sbmnl_fname = sbml_fname

        self._snapshot_sbml = re.sub(r'.xml$', '_snapshot.xml', sbml_fname)
        self._snapshot_json = re.sub(r'.xml$', '_snapshot.json', sbml_fname)

        if resume is True and os.path.exists(self._snapshot_sbml):
            sbml_fname = self._snapshot_sbml
        self.fba_model = FbaModel(sbml_fname)
        self.nrp = ProtectedParts(protected_parts)
        self.protected_sids = self.nrp.protected_sids

        if resume is True and os.path.exists(self._snapshot_json):
            with open(self._snapshot_json, 'r') as f:
                snapshot_data = json.load(f)
            self.protected_rids = set(snapshot_data['protected_rids'])
            print(f'Snapshot restored at {len(self.fba_model.rids):4d} remaining reactions '
                  f'from: {self._snapshot_json}')
        else:
            self.protected_rids = self.nrp.protected_rids
            # self.protected_rids = self.get_total_protected_rids()

        # overwrite the fba model bounds by general bound of protected parts (priot to LpProblem instantiation)
        self.fba_model.update_model_flux_bounds(self.nrp.overwrite_bounds)
        self.fba_model.update_model_reversible(self.get_pp_reversible_reactions())

        print('Determine optimal values for protected functions')
        self.set_function_optimum()
        print('Check consistency of protected parts')
        self.check_protected_parts()

    def get_total_protected_rids(self):
        """Determine protected reactions, incl. those leading to protected metabolites.

        :return: total protected reaction ids
        :rtype: set of str

        """
        # identify reactions that need protection due to protected metabolites
        #  over and above the protected rids provided in protected parts
        sid_mask = np.array([True if sid in self.protected_sids else False
                             for sid in self.fba_model.sids])
        psf_mask = self.fba_model.s_mat_coo.tocsr()[sid_mask].getnnz(axis=0)
        protected_sid_rids = set(self.fba_model.rids[psf_mask])
        protected_rids = self.nrp.protected_rids.union(protected_sid_rids)
        return protected_rids

    def get_pp_reversible_reactions(self):
        """Identify reversible reactions from protected parts

        :return: reaction ids from protected parts with negative fluxes
        :rtype: set of str
        """
        rev_pp_rids = set()
        for rid, bounds in self.nrp.overwrite_bounds.items():
            for idx in [0, 1]:
                if math.isfinite(bounds[idx]):
                    rev_pp_rids.add(rid)
        for pf in self.nrp.protected_functions.values():
            for rid, bounds in pf.overwrite_bounds.items():
                for idx in [0, 1]:
                    if math.isfinite(bounds[idx]) and bounds[idx] < 0.0:
                        rev_pp_rids.add(rid)
        return rev_pp_rids

    def set_function_optimum(self):
        """using FBA to set optimal value for protected functions"""
        for pf in self.nrp.protected_functions.values():
            res = self.fba_model.fba_pf_optimize(pf)
            pf.optimal = res['fun']
            pf.fba_success = res['success']
            if hasattr(pf, 'target_frac'):
                if pf.obj_dir == 'maximize':
                    pf.set_target_lb(pf.target_frac * pf.optimal)
                else:
                    pf.set_target_ub(pf.target_frac * pf.optimal)

    def check_protected_parts(self):
        """Report on consisteny issues wrt to protected parts.

        E.g. report on reactions and metabolties that do not exist in the model
        """
        extra_p_rids = self.protected_rids.difference(set(self.fba_model.rids))
        extra_p_sids = self.protected_sids.difference(set(self.fba_model.sids))
        pf_rids = set()
        for pf in self.nrp.protected_functions.values():
            pf_rids |= {rid for rid in pf.objective}
            pf_rids |= {rid for rid in pf.overwrite_bounds}
            if pf.fba_success is False:
                print(f'Protected function {pf.obj_id} has unsuccessful FBA '
                      f' with optimal value {pf.optimal}')

        extra_pf_rids = pf_rids.difference(set(self.fba_model.rids))
        pf_unprotected_rids = pf_rids.difference(set(self.protected_rids))
        if len(extra_p_rids) > 0:
            print('Protected reactions that do not exist in model:', extra_p_rids)
        if len(extra_p_sids) > 0:
            print('Protected metabolites that do not exist in model:', extra_p_sids)
        if len(extra_pf_rids) > 0:
            print('Reactions in protected functions that do not exist in model:', extra_pf_rids)
        if len(pf_unprotected_rids) > 0:
            print('Reactions in protected functions that are not protected:', pf_unprotected_rids)

        blocked_rids = self.reaction_types_fva(self.fba_model.rids)['blocked_rids']
        blocked_p_rids = self.nrp.protected_sids.intersection(blocked_rids)
        blocked_pf_rids = pf_rids.intersection(blocked_rids)
        if len(blocked_p_rids) > 0:
            print('Protected reactions that are blocked (no flux):', blocked_p_rids)
        if len(blocked_pf_rids) > 0:
            print('Reactions in protected functions that are blocked (no flux):', blocked_pf_rids)

    def identify_parallel_reactions(self):
        """Identify parallel reactions having same stoichiometry (considering reverse).

        protected reaction are preserved.
        reversible reactions have higher priority

        :return: reaction ids of parallel reactions
        :rtype: set of str
        """
        drop_rids = set()
        for group_rids in self.fba_model.get_parallel_rids():
            protected = set()
            rid_rev = None
            for rid in group_rids:
                if rid in self.protected_rids:
                    protected.add(rid)
                if rid in self.fba_model.rids_rev:
                    rid_rev = rid
            if len(protected) > 0:
                drop_group_rids = set(group_rids) - protected
            elif rid_rev is not None:
                drop_group_rids = set(group_rids) - {rid_rev}
            else:
                drop_group_rids = set(group_rids[1:])
            drop_rids |= drop_group_rids
        return drop_rids

    def reaction_types_fva(self, free_rids):
        """Idenify reaction types using FVA applied to all protected functions

        first part: run FVA across all conditions

        :param free_rids: reaction ids used for FVA
        :type free_rids: list of strings
        :return: rid, min and max flux values
        :rtype: dict
        """
        print(time.strftime("%H:%M:%S", time.localtime()))

        n_pf = len(self.nrp.protected_functions)
        flux_min_pfs = np.zeros((n_pf, len(free_rids)))
        flux_max_pfs = np.zeros((n_pf, len(free_rids)))

        res = {}
        # setting up multiprocessing takes quite some resources
        #    there is a tradeoff wrt running on one processor
        processes = min(self.cpus, int(len(free_rids)*n_pf/self.min_mp_total_flux))
        # processes = min(self.cpus, len(free_rids))

        if processes > 1:
            print(f'Multiprocessing with {processes} processes')
            frid2idx = {rid: idx for idx, rid in enumerate(free_rids)}
            # delete lp problem prior to pool creation (so FbaProblem can be pickled)
            self.fba_model.delete_fba_lp()
            with mp.Pool(processes, initializer=_init_worker,
                         initargs=(self.fba_model, self.nrp.protected_functions)) as pool:
                for res in pool.imap_unordered(_worker, np.array_split(free_rids, processes)):
                    idxs = [frid2idx[rid] for rid in res['rids']]
                    flux_min_pfs[:, idxs] = res['min_pfs']
                    flux_max_pfs[:, idxs] = res['max_pfs']
        else:
            for idx, pf in enumerate(self.nrp.protected_functions.values()):
                # print(f'\nCheck FVA for pf {pf.obj_id} with {len(free_rids)} free variables')
                res = self.fba_model.fva_pf_flux_ranges(pf, free_rids)
                if res['success'] is True:
                    flux_min_pfs[idx] = res['min']
                    flux_max_pfs[idx] = res['max']
                else:
                    print(f'FBA for condition {pf.obj_id} is not feasible')
                    # TODO: can this happen? how best to continue?
                    flux_min_pfs[idx] = np.nan
                    flux_max_pfs[idx] = np.nan
                    break
                sys.stdout.write('\rFVA[{}{}] {:3.0f}%'.format(
                    '=' * int(idx+1), ' ' * int(n_pf - idx-1), (idx+1)/n_pf*100))
                sys.stdout.flush()

        if res['success'] is False:
            reaction_types = {'blocked_rids': set(), 'essential_rids': set(free_rids),
                              'candidate_rids_sorted': []}
        else:
            reaction_types = self.classify_reactions(free_rids, flux_min_pfs, flux_max_pfs)
        return reaction_types

    def classify_reactions(self, rids, flux_min_pfs, flux_max_pfs):
        """Classify reaction types based on FVA results.

        determine flux span (min/max flux) for each reaction across all conditions
        Identify blocked, essential and deletion candidate reactions based on flux span
            blocked reactions: min and max of flux span is zero
            essential reactions: min and max either > 0 or < 0
            deletion candidates: flux span support zero (0) flux

        :param rids: reaction ids used in FVA
        :type rids: list of strings
        :param flux_min_pfs: minimum fluxes per condition
        :type flux_min_pfs: 2D ndarray of floats (reactions x conditions)
        :param flux_max_pfs: maximum fluxes per condition
        :type flux_max_pfs: 2D ndarray of floats (reactions x conditions)
        :return: rid, min and max flux values
        :rtype: dict
        """
        # analyze flux ranges per reaction, aggregate fluxes
        flux_min_pfs[np.abs(flux_min_pfs) < self.tolerance] = 0.0
        flux_max_pfs[np.abs(flux_max_pfs) < self.tolerance] = 0.0
        flux_min = np.min(flux_min_pfs, axis=0)
        flux_max = np.max(flux_max_pfs, axis=0)

        # blocked reactions do not carry any flux in all conditions
        mask_b = np.all(np.vstack((flux_min == 0, flux_max == 0)), axis=0)
        blocked_rids = set(np.array(rids)[mask_b])

        # essential reactions carry either only forward or reverse flux, exluding zero flux
        mask_e1 = np.all(np.vstack((flux_min < 0, flux_max < 0)), axis=0)
        mask_e2 = np.all(np.vstack((flux_min > 0, flux_max > 0)), axis=0)
        mask_e = np.any(np.vstack((mask_e1, mask_e2)), axis=0)
        essential_rids = set(np.array(rids)[mask_e])

        # reactions that support zero flux are considered candiates for removal
        #  here such reactions are sorted from smallest to largest flux span
        mask_c = np.all(np.vstack((np.logical_not(mask_b), np.logical_not(mask_e))), axis=0)
        flux_span = flux_max - flux_min
        ind = np.argsort(flux_span)
        candidate_rids_sorted = list(np.array(rids)[ind][np.array(mask_c)[ind]])

        reaction_types = {'blocked_rids': blocked_rids, 'essential_rids': essential_rids,
                          'candidate_rids_sorted': candidate_rids_sorted}

        return reaction_types

    def prune(self, snapshot_freq=0):
        """Prune the metabolic network

        Once the Metabolic network has been loaded it can be pruned.
        With snapshot_freq one can deterimine if (snapshot_freq > 0) and
        how often intermediate results shall be store, from which one can resume in future

        Step during Pruning
        1. identification and removal of parallel reactions
        2. check feasibility of the network wrt to all protected functions
        3. in a loop till all free reactions have been consumed
        3.1 identify types of remaining free reactions using FVA across all protected functions
        3.2 blocked reactions (zero flux across all functions) get removed
        3.3 essential reactions (strict positive/negative flux across conditions) get protected
        3.4 a single candidate reaction (supporting zero flux) gets identify and removed

        :param snapshot_freq: frequency to store intermediate results (every x reactions)
        :type snapshot_freq: int, optional (default: 0, no snapshots)
        """

        print(f'Initial S-matrix: {self.fba_model.s_mat_coo.shape}; '
              f' protected reactions: {len(self.protected_rids)}:', self.protected_rids)
        drop_rids = self.identify_parallel_reactions().difference(self.protected_rids)
        if len(drop_rids) > 0:
            self.fba_model.remove_reactions(drop_rids)
            print(f'{len(drop_rids)} parallel reaction(s) dropped:', drop_rids)

        feasible = np.array([self.fba_model.check_pf_feasibility(pf) for pf in self.nrp.protected_functions.values()])
        print('Protected functions feasibility:', np.all(feasible))

        next_snapshot = 0
        if snapshot_freq > 0:
            next_snapshot = max(0, len(self.fba_model.rids) - snapshot_freq)

        while len(self.protected_rids) < self.fba_model.s_mat_coo.shape[1]:
            free_rids = list(set(self.fba_model.rids).difference(self.protected_rids))
            reaction_types = self.reaction_types_fva(free_rids)
            print()
            blocked_rids = reaction_types['blocked_rids']
            essential_rids = reaction_types['essential_rids']
            candidate_rids = reaction_types['candidate_rids_sorted']
            if len(blocked_rids) > 0:
                self.fba_model.remove_reactions(blocked_rids)
                print(f"{len(blocked_rids)} blocked reaction(s) removed:", blocked_rids)
            print(f"{len(candidate_rids)} candidates:", candidate_rids[:4], '...')

            # try to remove one reaction while still satisfying all protected functions
            for rid in candidate_rids:
                # print('Check feasibility of', rid)
                feasible = True
                for pf in self.nrp.protected_functions.values():
                    feasible = self.fba_model.check_pf_feasibility(pf, zero_flux_rid=rid)
                    # print(f'Feasibility of {rid} is {feasible}')
                    if feasible is False:
                        essential_rids.add(rid)
                        break
                if feasible is True:
                    self.fba_model.remove_reactions({rid})
                    print(f'{rid} removed from the model')
                    break

            if len(essential_rids) > 0:
                self.protected_rids |= essential_rids
                print(f"{len(essential_rids)} reaction(s) added to protected reactions:", essential_rids)

            if feasible is False:
                print('No more reactions to remove')

            # store intermediate results (snapshots)
            if len(self.fba_model.rids) < next_snapshot:
                self.export_pruned_model(self._snapshot_sbml)
                snapshot_data = {'protected_rids': list(self.protected_rids)}
                with open(self._snapshot_json, 'w') as f:
                    json.dump(snapshot_data, f)
                next_snapshot = max(0, len(self.fba_model.rids) - snapshot_freq)
                print('snapshot data saved.')

            print(f'S-matrix: {self.fba_model.s_mat_coo.shape}; '
                  f'protected reactions: {len(self.protected_rids)}')

        # check if blocked reactions exist in the finally pruned model
        # blocked_rids = self.reaction_types_fva(self.fba_model.rids)['blocked_rids']
        # if len(blocked_rids) > 0:
        #     print(f'{len(blocked_rids)} blocked reaction(s) in pruned model', blocked_rids)

    def remove_snapshot_files(self):
        """Remove any snapshot files still existing"""
        for pname in [self._snapshot_sbml, self._snapshot_json]:
            if os.path.exists(pname):
                os.remove(pname)
                print(f'{pname} removed')

    def get_reduction_status(self):
        """ Return status in model reduction process

        degrees of freedom, number removed reactions, number essential reactions
        number remaining reactions
        :returns: status information in a dictionary
        :rtype: dict
        """
        status = self.fba_model.get_model_status()
        status['n_protected_rids'] = len(self.protected_rids)
        status['n_protected_sids'] = len(self.protected_sids)
        return status

    def print_status(self):
        """Print status of pruning process"""

        status = self.get_reduction_status()
        print("reactions removed/remaining/protected: "
              f"{status['n_full_rids' ] - status['n_rids']:4d}/{status['n_rids']:4d}"
              f"/{len(self.protected_rids):4d}; dof: {status['dof']:3d}")

    def export_pruned_model(self, pruned_sbml=None):
        """Export the pruned network (or snapshot) to SBML.

        Component information of initial network will be carried over,
        e.g. annotation data, groups, gene products

        :param pruned_sbml: pathname of pruned network, ending with '.xml'
        :type pruned_sbml: str, optional (default: None, construct a name)
        """
        n_s, n_r = self.fba_model.s_mat_coo.shape
        if pruned_sbml is None:
            pruned_sbml = re.sub(r'.xml$', f'_pruned_{n_s}x{n_r}.xml', self.full_sbmnl_fname)

        success = self.fba_model.export_pruned_model(pruned_sbml)
        if success is True:
            print('Pruned model exported to', pruned_sbml)
            if len(self.protected_rids) == n_r:
                self.remove_snapshot_files()
        else:
            print('Pruned model could not be exported')

    def __del__(self):
        """Delete corresponding LpProblem"""
        del self.fba_model
