
import os
import time
import math
import numpy as np
import pandas as pd
import sbmlxdf


class ProtectedParts:

    def __init__(self, protected_parts):

        if type(protected_parts) == str:
            if os.path.exists(protected_parts):
                nrp = self.load_raw_protected_data(protected_parts)
                print(f'Protected data loaded from {protected_parts} '
                      f'(last modified: {time.ctime(os.path.getmtime(protected_parts))})')
            else:
                print(f'{protected_parts} not found!')
                raise FileNotFoundError
        else:
            assert type(protected_parts) == dict, 'expected filename or a dict'
            nrp = protected_parts

        self.protected_rids = set(nrp.get('reactions', []))
        self.protected_sids = set(nrp.get('metabolites', []))
        if 'bounds' in nrp:
            num_cols = ['fbcLb', 'fbcUb']
            nrp['bounds'][num_cols] = nrp['bounds'][num_cols].applymap(
                lambda x: np.nan if type(x) == str else x)
            self.overwrite_bounds = {rid: [row['fbcLb'], row['fbcUb']]
                                     for rid, row in nrp['bounds'].iterrows()}
        else:
            self.overwrite_bounds = {}

        self.protected_functions = {}
        if 'functions' in nrp:
            num_cols = ['fbcLb', 'fbcUb', 'fraction']
            nrp['functions'][num_cols] = nrp['functions'][num_cols].applymap(
                lambda x: np.nan if type(x) == str else x)
            str_cols = ['fluxObjectives', 'reaction']
            nrp['functions'][str_cols] = nrp['functions'][str_cols].applymap(
                lambda x: '' if type(x) != str else x)
            for func_id in list(nrp['functions'].index.unique()):
                df_pf = nrp['functions'].loc[func_id]
                self.protected_functions[func_id] = ProtectedFunction(df_pf)

    @staticmethod
    def load_raw_protected_data(fname):
        """Load raw protected parts from Excel file.

        Reaction/metabolite identifiers not modified.

        :param fname: file name with protected parts
        :type fname: str
        :returns: protected parts including:
        :rtype: dict
        """
        nrp = {}
        with pd.ExcelFile(fname) as xlsx:
            for key in ['reactions', 'metabolites']:
                if key in xlsx.sheet_names:
                    nrp[key] = pd.read_excel(xlsx, sheet_name=key).iloc[:, 0].to_list()
            for key in ['functions', 'bounds']:
                if key in xlsx.sheet_names:
                    nrp[key] = pd.read_excel(xlsx, sheet_name=key, index_col=0)
        return nrp


class ProtectedFunction:

    def __init__(self, df_pf):
        if type(df_pf) == pd.Series:
            df_pf = pd.DataFrame([df_pf.values], columns=df_pf.index, index=[df_pf.name])

        self.obj_id = df_pf.index[0]
        self.objective = {}
        self.overwrite_bounds = {}
        for _, row in df_pf.iterrows():
            if len(row['fluxObjectives']) > 0:
                self.obj_dir = row['type']
                for record in sbmlxdf.misc.record_generator(row['fluxObjectives']):
                    params = sbmlxdf.misc.extract_params(record)
                    self.objective[params['reac']] = float(params['coef'])
                if math.isfinite(row['fraction']):
                    self.target_frac = row['fraction']
                if math.isfinite(row['fbcLb']):
                    self.target_lb = row['fbcLb']
                if math.isfinite(row['fbcUb']):
                    self.target_ub = row['fbcUb']
            if len(row.reaction) > 0:
                self.overwrite_bounds[row['reaction']] = [row['fbcLb'], row['fbcUb']]
            if hasattr(row, 'notes'):
                if type(row['notes']) == str and len(row['notes']) > 0:
                    self.notes = row['notes']

    def set_target_lb(self, lb):
        self.target_lb = lb

    def set_target_ub(self, ub):
        self.target_ub = ub
