

import os
import time
import re
import math
import numpy as np
from scipy import sparse
import sbmlxdf

from modelpruner.problems.glpk_linear_problem import GlpkLinearProblem


class FbaModel:

    def __init__(self, sbml_fname):

        self.sbml_fname = sbml_fname
        if os.path.exists(sbml_fname):
            sbml_model = sbmlxdf.Model(sbml_fname)
            print(f'SBML model loaded from {sbml_fname} '
                  f'(last modified: {time.ctime(os.path.getmtime(sbml_fname))})')
        else:
            print(f'{sbml_fname} not found!')
            raise FileNotFoundError

        model_dict = sbml_model.to_df()
        df_reactions = model_dict['reactions']
        df_species = model_dict['species']

        self.rids = df_reactions.index.to_numpy()
        self.rid2idx = {rid: idx for idx, rid in enumerate(self.rids)}
        self.sids = df_species.index.to_numpy()
        self.sid2idx = {sid: idx for idx, sid in enumerate(self.sids)}
        self.full_model_shape = {'n_full_rids': len(self.rids), 'n_full_sids': len(self.sids)}
        self.flux_bounds = np.vstack((df_reactions['fbcLb'].to_numpy(),
                                      df_reactions['fbcUb'].to_numpy()))
        self.reversible = df_reactions['reversible'].to_numpy().copy()
        for idx in range(len(self.reversible)):
            if self.flux_bounds[0, idx] < 0 and np.logical_not(self.reversible[idx]):
                self.reversible[idx] = True
                print(f'reaction {idx} set to reversible')
        self.rids_rev = self.rids[self.reversible]
        self.rid2idx_rev = {rid: idx + len(self.rids) for idx, rid in enumerate(self.rids_rev)}
        # using a sparse matrix speeds up significantly pickling (factor > 350 for S-matrix)
        # pickling required to support multiprocessing
        # also data exchange from sbmlxdf.Model and towards LpProblem improves
        self.s_mat_coo = sparse.coo_array(sbml_model.get_s_matrix(sparse=True))

        # lp problem will only be created once required.
        # this gives chances to update e.g. flux bounds and to support pickling (multiprocessing)
        self.lp = None

    def update_reversible_rids(self):
        self.rids_rev = self.rids[self.reversible]
        self.rid2idx_rev = {rid: idx + len(self.rids) for idx, rid in enumerate(self.rids_rev)}

    def update_model_reversible(self, reactions):
        """Update reversible flag of reactions

        :param reactions: reactions ids that have negative flux bounds
        :type reactions: set of str
        """
        upd_rids = set(reactions).difference(set(self.rids_rev))
        for rid in upd_rids:
            self.reversible[self.rid2idx[rid]] = True
        if len(upd_rids) > 0:
            print('reactions in protected functions made reversible:', upd_rids)
            self.update_reversible_rids()

    def update_model_flux_bounds(self, flux_bounds):
        """Overwrite model flux bounds with protected parts flux bounds.

        lp problem might need to be instantiated

        :param flux_bounds: lower and upper flux bounds for selected reactions
        :type flux_bounds: dict: key: reaction id, value: list-like of two float values
        """
        for rid, bound_upd in flux_bounds.items():
            for i in [0, 1]:
                if math.isfinite(bound_upd[i]):
                    self.flux_bounds[i, self.rid2idx[rid]] = bound_upd[i]

    def get_parallel_rids(self):
        """Identify identical reactions (forward/reverse).

        Linerar scaling is not considered.

        :return: lists of reaction ids that are parallel
        :rtype: list with lists of str
        """

        fwd_rev_s_mat_coo = sparse.hstack([self.s_mat_coo,
                                           -self.s_mat_coo.tocsc()[:, self.reversible]])
        fwd_rev_rids = np.hstack((self.rids, self.rids_rev))

        # identify groups of parallel reactions
        sorted_idx = np.lexsort(fwd_rev_s_mat_coo.todense())[0]
        groups = []
        consumed_idxs = set()
        for pos in range(len(sorted_idx)):
            idx = sorted_idx[pos]
            if idx not in consumed_idxs:
                group = [idx]
                for adjacent_pos in range(pos + 1, len(sorted_idx)):
                    candiate_idx = sorted_idx[adjacent_pos]
                    if (fwd_rev_s_mat_coo.getcol(idx) != fwd_rev_s_mat_coo.getcol(candiate_idx))\
                            .getnnz() == 0:
                        consumed_idxs.add(candiate_idx)
                        group.append(candiate_idx)
                    else:
                        break
                if len(group) > 1:
                    groups.append(group)

        parallel_rid_groups = [list(fwd_rev_rids[group]) for group in groups]
        return parallel_rid_groups

    def combine_fwd_rev(self, vector):
        """combine forward and backward flux information.

        :param vector: contains data on all reactions + backward information for reversible reactions
        :type vector: 1D ndarray of shape (len(mr_rids) + sum(reversible))
        :return: combined flux information
        :rtype: 1D ndarray of shape (len(mr_rids))
        """
        n_cols = len(self.rids)
        combined = vector[:n_cols]
        combined[self.reversible] -= vector[n_cols:]
        return combined

    def get_model_status(self):
        """ Return status on model


        :returns: status information in a dictionary
        :rtype: dict
        """
        status = {'dof': self.s_mat_coo.shape[1] - np.linalg.matrix_rank(self.s_mat_coo.todense()),
                  'n_rids': len(self.rids),
                  'n_sids': len(self.sids)}
        status.update(self.full_model_shape)
        return status

    def export_pruned_model(self, pruned_sbml):
        """Export the pruned network (or snapshot) to SBML.

        Component information of initial network will be carried over,
        e.g. annotation data, groups, gene products
        History information, if available, will be updated.

        Model will also be validated with respect to SBML compliance.

        :param pruned_sbml: pathname of pruned network, ending with '.xml'
        :type pruned_sbml: str
        """
        sbml_model = sbmlxdf.Model(self.sbml_fname)
        model_dict = sbml_model.to_df()
        pruned_mdict = model_dict.copy()
        pruned_mdict['reactions'] = model_dict['reactions'].loc[self.rids]
        pruned_mdict['species'] = model_dict['species'].loc[self.sids]

        # update model attributes (ids, names and modification date)
        pruned_mdict['modelAttrs']['id'] += '_pruned'
        if 'name' in pruned_mdict['modelAttrs']:
            pruned_mdict['modelAttrs']['name'] += '_pruned'
        if 'metaid' in pruned_mdict['modelAttrs']:
            pruned_mdict['modelAttrs']['metaid'] += '_pruned'
        if 'modifiedHistory' in pruned_mdict['modelAttrs']:
            pruned_mdict['modelAttrs']['modifiedHistory'] += '; localtime'

        # clean groups table
        keep_idxs = []
        if 'groups' in pruned_mdict:
            df_groups_pruned = pruned_mdict['groups'].copy()
            for idx, row in df_groups_pruned.iterrows():
                members_pruned = []
                for member in sbmlxdf.misc.record_generator(row['members']):
                    params = sbmlxdf.misc.extract_params(member)
                    if 'idRef' in params:
                        if params['idRef'] in self.rids:
                            members_pruned.append(member)
                    else:
                        members_pruned.append(member)
                df_groups_pruned.at[idx, 'members'] = '; '.join(members_pruned)
                if len(members_pruned) > 0:
                    keep_idxs.append(idx)
            pruned_mdict['groups'] = df_groups_pruned.loc[keep_idxs]

        # clean gene products
        gp_ids = set()
        if 'fbcGeneProdAssoc' in pruned_mdict['reactions']:
            for gpr in pruned_mdict['reactions']['fbcGeneProdAssoc']:
                if type(gpr) is str:
                    if 'assoc=' in gpr:
                        gpr = re.sub('assoc=', '', gpr)
                        gpr = re.sub(r'[()]', ' ', gpr)
                        gpr = re.sub(r' and ', ' ', gpr)
                        gpr = re.sub(r' or ', ' ', gpr)
                    gp_ids = gp_ids.union(set(re.findall(r'\w+', gpr)))
        if 'fbcGeneProducts' in pruned_mdict:
            pruned_mdict['fbcGeneProducts'] = pruned_mdict['fbcGeneProducts'].loc[list(gp_ids)]

        pruned_model = sbmlxdf.Model()
        pruned_model.from_df(pruned_mdict)
        err = pruned_model.validate_sbml()
        if len(err) > 0:
            print('SBML validation result:', ', '.join([k + ': ' + str(v) for k, v in err.items()]))
            print('see ./results/tmp.txt for details')
        return pruned_model.export_sbml(pruned_sbml)

    def create_core_fba_lp(self):
        """create a core Linear Programming Problem for FBA with split reactions

        Core Problem contains coefficient matrix, variable and row bounds
        Objective function yet needs to be added prior to any optimization

        Splitting reactions in forward/backward directions reduces high values during FBA
        configured bounds are >= 0
        Splitting required for pFBA (minimize absolute flux values)

        :return: LpProblem configured for FBA
        :rtype: LpProblem
        """
        fwd_rev_s_mat_coo = sparse.hstack([self.s_mat_coo,
                                           -self.s_mat_coo.tocsc()[:, self.reversible]])
        fwd_rev_col_bnds = np.hstack((np.fmax(self.flux_bounds[:], 0),
                                      np.fmax(np.flip(-self.flux_bounds[:, self.reversible], axis=0), 0)))
        row_bnds = np.zeros((fwd_rev_s_mat_coo.shape[0], 2))
        lp = GlpkLinearProblem(fwd_rev_s_mat_coo, row_bnds, fwd_rev_col_bnds)
        return lp

    def delete_fba_lp(self):
        """Delete the corresponding linear problem.

        E.g. required for multiprocessing so FbaModel can be pickled
        """
        if self.lp is not None:
            del self.lp
            self.lp = None

    def set_lp_objective(self, objective, direction='maximize'):
        """Set objective coefficients and direction in linear problem

        :param objective: objective coefficients
        :type objective: dict with key: reaction id, value: stoichiometric coefficient
        :param direction: direction of optimization
        :type direction: str (optional, default: 'maximize')
        """
        assert self.lp is not None
        self.lp.set_obj_coefs(self.get_coefficients(objective))
        self.lp.set_obj_dir(direction)

    def remove_lp_objective(self, objective):
        """remove objective coefficients

        :param objective: objective coefficients
        :type objective: dict with key: reaction id, value: stoichiometric coefficient
        """
        self.lp.remove_obj_coefs(self.get_coefficients(objective))

    def update_lp_flux_bounds(self, flux_bounds):
        """update flux bounds for selected reactions in the linear problem.

        lp problem might need to be instantiated
        overwrite bounds of protected functions will be cleaned from already removed reactions

        :param flux_bounds: lower and upper flux bounds for selected reactions
        :type flux_bounds: dict: key: reaction id, value: list-like of two float values
        """
        if self.lp is None:
            self.lp = self.create_core_fba_lp()
            # print('Core LP Problem created')
        for rid, bound_upd in flux_bounds.items():
            if rid in self.rids:
                temp_bounds = self.flux_bounds[:, self.rid2idx[rid]].copy()
                for i in [0, 1]:
                    if math.isfinite(bound_upd[i]):
                        temp_bounds[i] = bound_upd[i]
                self.lp.set_col_bnd(self.rid2idx[rid], np.fmax(temp_bounds, 0))
                if rid in self.rids_rev:
                    self.lp.set_col_bnd(self.rid2idx_rev[rid], np.fmax(np.flip(-temp_bounds), 0))

    def reset_lp_flux_bounds(self, rids):
        """restore original flux bounds for selected reactions in linear problem.

        overwrite bounds of protected functions will be cleaned from already removed reactions

        :param rids: reaction ids for which to restore the model orignal flux bounds
        :type rids: set of str
        """
        assert self.lp is not None

        for rid in rids:
            if rid in self.rids:
                assert rid in self.rids
                bounds = self.flux_bounds[:, self.rid2idx[rid]]
                self.lp.set_col_bnd(self.rid2idx[rid], np.fmax(bounds, 0))
                if rid in self.rids_rev:
                    self.lp.set_col_bnd(self.rid2idx_rev[rid], np.fmax(np.flip(-bounds), 0))

    def get_coefficients(self, objective):
        """get coefficients for objective or constraint function

        :param objective: objective/constraint coefficients
        :type objective: dict with key: reaction id, value: stoichiometric coefficient
        :return: coefficients aligned with the linear problem
        :rtype: 1D ndarray with float
        """
        coefs = np.zeros(len(self.rids) + len(self.rids_rev))
        for rid, coef in objective.items():
            coefs[self.rid2idx[rid]] = coef
            if rid in self.rids_rev:
                coefs[self.rid2idx_rev[rid]] = -coef
        return coefs

    def remove_reactions(self, drop_rids):
        """remove reactions and blocked metabolites from model and lp problem

        :param drop_rids: reaction ids to be removed from the model
        :type drop_rids: set of str
        """
        if self.lp is None:
            self.lp = self.create_core_fba_lp()
            # print('Core LP Problem created')
        assert self.lp is not None

        # remove variables (columns) from linear problem
        col_idxs0_fwd = [self.rid2idx[rid] for rid in drop_rids]
        col_idxs0_rev = [self.rid2idx_rev[rid] for rid in drop_rids if rid in self.rids_rev]
        self.lp.remove_cols(col_idxs0_fwd + col_idxs0_rev)

        keep_rids_mask = np.array([False if rid in drop_rids else True for rid in self.rids])
        self.rids = self.rids[keep_rids_mask]
        self.rid2idx = {rid: idx for idx, rid in enumerate(self.rids)}
        self.reversible = self.reversible[keep_rids_mask]
        self.update_reversible_rids()
        self.flux_bounds = self.flux_bounds[:, keep_rids_mask]
        self.s_mat_coo = self.s_mat_coo.tocsc()[:, keep_rids_mask].tocoo()

        # drop blocked metabolites due to deleted reactions
        keep_sids_mask = (self.s_mat_coo.getnnz(axis=1) > 0)
        row_idxs0 = [idx for idx, val in enumerate(keep_sids_mask) if val == np.bool_(False)]
        if len(row_idxs0) > 0:
            self.lp.remove_rows(row_idxs0)
            self.sids = self.sids[keep_sids_mask]
            self.sid2idx = {sid: idx for idx, sid in enumerate(self.sids)}
            self.s_mat_coo = self.s_mat_coo.tocsr()[keep_sids_mask, :].tocoo()

    def fba_pf_optimize(self, pf, zero_flux_rid=None):
        """FBA optimization of current problem

        zero_flux_rids overwrites settings in protected functions

        :param pf: protected function (objective)
        :type pf: class ProtectedFunction
        :param zero_flux_rid: reaction id of candiated checked for removal
        :type zero_flux_rid: str, optional (default: None)
        :return: results information from the optimization
        :rtype: dict
        """
        if self.lp is None:
            self.lp = self.create_core_fba_lp()
            # print('Core LP Problem created')
        assert self.lp is not None

        if zero_flux_rid is not None:
            overwrite_bounds = pf.overwrite_bounds.copy()
            overwrite_bounds[zero_flux_rid] = np.array([0.0, 0.0])
        else:
            overwrite_bounds = pf.overwrite_bounds

        self.set_lp_objective(pf.objective, pf.obj_dir)
        self.update_lp_flux_bounds(overwrite_bounds)
        res = self.lp.solve()
        # print(res['success'], res['fun'])
        self.reset_lp_flux_bounds(overwrite_bounds.keys())
        self.remove_lp_objective(pf.objective)
        return res

    def check_pf_feasibility(self, pf, zero_flux_rid=None):
        """check that current model satisfies the protected functions

        :param pf: protected function (objective)
        :type pf: class ProtectedFunction
        :param zero_flux_rid: reaction id of candiated checked for removal
        :type zero_flux_rid: str, optional (default: None)
        :return: status if protected function satisfied
        :rtype: bool
        """
        feasible = True
        res = self.fba_pf_optimize(pf, zero_flux_rid)
        if res['success'] is False:
            feasible = False
        elif pf.obj_dir == 'maximize':
            if res['fun'] < pf.target_lb:
                feasible = False
        else:
            if res['fun'] > pf.target_ub:
                feasible = False
        return feasible

    def fva_pf_flux_ranges(self, pf, rids):
        """FVA for a specific protected function for all or selected reactions

        Reference: Mahadevan R, Schilling CH: The effects of alternate optimal
        solutions in constraint-based genome-scale metabolic models.
        Metab Eng 2003, 5(4):264-76.

        :param pf: protected function (objective)
        :type pf: class ProtectedFunction
        :param rids: list of reaction ids to get flux range
        :type rids: list of str
        :return: rid, min and max flux values
        :rtype: dict
        """
        # check that problem is feasible
        if self.check_pf_feasibility(pf) is False:
            return {'success': False}

        # implement objective as a constraint and overwrite bounds
        row_bnds = (np.array([pf.target_lb, np.nan]) if pf.obj_dir == 'maximize'
                    else np.array([np.nan, pf.target_ub]))
        row_idx0 = self.lp.add_constraint(self.get_coefficients(pf.objective), row_bnds)
        self.update_lp_flux_bounds(pf.overwrite_bounds)

        flux_min = np.zeros(len(rids))
        flux_max = np.zeros(len(rids))
        for idx, rid in enumerate(rids):
            self.set_lp_objective({rid: 1}, 'minimize')
            res = self.lp.solve()
            flux_min[idx] = res['fun'] if res['success'] else np.nan
            self.lp.set_obj_dir('maximize')
            res = self.lp.solve()
            flux_max[idx] = res['fun'] if res['success'] else np.nan
            self.remove_lp_objective({rid: 1})

        self.reset_lp_flux_bounds(pf.overwrite_bounds.keys())
        self.lp.del_constraint(row_idx0)

        res = {'success': True, 'rids': np.array(rids), 'min': flux_min, 'max': flux_max}
        # print(res)
        return res

    def copy(self):
        """Copy the FbaModel instance (while removing linear problem).

        A shallow copy, to support multiprocessing of FVA,
        which do not impact FBA model attributes, except for linear problem
        """
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        result.__dict__['lp'] = None
        return result

    def __del__(self):
        print('deletion of fba_model')
        del self.lp
