"""Implementation of LpProblem (linear programming problem) class.

built upon swiglpk

Peter Schubert, HHU Duesseldorf, June 2022
"""

import numpy as np
import swiglpk as glpk


# status reported
simplex_status = {
    0: 'LP problem instance successfully solved',
    glpk.GLP_EBADB: 'Unable to start search, initial basis specified in the '
       'problem object is invalid. Number basic variables is not same as the number rows.',
    glpk.GLP_ESING: 'Unable to start search, basis matrix corresponding to initial basis is singular.',
    glpk.GLP_ECOND: 'Unable to start search, basis matrix corresponding to initial basis is ill-conditioned.',
    glpk.GLP_EBOUND: 'Unable to start search, some double-bounded variables have incorrect bounds.',
    glpk.GLP_EFAIL: 'Search prematurely terminated: solver failure.',
    glpk.GLP_EOBJLL: 'Search prematurely terminated: objective function being '
       'maximized reached its lower limit and continues decreasing.',
    glpk.GLP_EOBJUL: 'Search prematurely terminated: objective function being '
       'minimized reached its upper limit and continues increasing.',
    glpk.GLP_EITLIM: 'Search prematurely terminated: simplex iteration limit exceeded.',
    glpk.GLP_ETMLIM: 'Search prematurely terminated: time limit exceeded.',
    glpk.GLP_ENOPFS: 'LP problem instance has no primal feasible solution.',
    glpk.GLP_ENODFS: 'LP problem instance has no dual feasible solution.'}

generic_status = {glpk.GLP_UNDEF: 'solution is undefined',
                  glpk.GLP_FEAS: 'solution is feasible',
                  glpk.GLP_INFEAS: 'solution is infeasible',
                  glpk.GLP_NOFEAS: 'problem has no feasible solution',
                  glpk.GLP_OPT: 'solution is optimal',
                  glpk.GLP_UNBND: 'problem has unbounded solution'}

dual_status = {glpk.GLP_UNDEF: 'dual solution is undefined',
               glpk.GLP_FEAS: 'dual solution is feasible',
               glpk.GLP_INFEAS: 'dual solution is infeasible',
               glpk.GLP_NOFEAS: 'no dual feasible solution exists'}

prim_status = {glpk.GLP_UNDEF: 'primal solution is undefined',
               glpk.GLP_FEAS: 'primal solution is feasible',
               glpk.GLP_INFEAS: 'primal solution is infeasible',
               glpk.GLP_NOFEAS: 'no primal feasible solution exists'}

var_status = {glpk.GLP_BS: 'basic variable',
              glpk.GLP_NL: 'non-basic variable on its lower bound',
              glpk.GLP_NU: 'non-basic variable on its upper bound',
              glpk.GLP_NF: 'non-basic free (unbounded) variable',
              glpk.GLP_NS: 'non-basic fixed variable'}

scaling_options = {'GM': glpk.GLP_SF_GM,
                   'EQ': glpk.GLP_SF_EQ,
                   '2N': glpk.GLP_SF_2N,
                   'SKIP': glpk.GLP_SF_SKIP,
                   'AUTO': glpk.GLP_SF_AUTO}


class GlpkLinearProblem:

    def __init__(self, constr_coefs_coo, row_bnds, col_bnds):
        """create core linear problem (constraints, variable and row bounds)

        Problem needs to be properly deleted usin del(LpProblem)
        Note: glpk related indices start at 1 (not zero)
        :param constr_coefs_coo: matrix with constraint coefficients
        :type constr_coefs_coo: scipy.sparce.coo_matrix with shape(nrows, ncols)
        :param row_bnds: lower and upper bounds of auxhiliary (row) variables
        :type row_bnds: 2D ndarray of shape(nrows, 2) (1st col: lower bounds, 2nd col: upper bounds)
        :param col_bnds: lower and upper bounds of structural (col) variables
        :type col_bnds: 2D ndarray of shape(2, ncols) (1st row: lower bounds, 2nd row: upper bounds)
        """
        glp_msg_lev = glpk.GLP_MSG_OFF    # GLP_MSG_OFF, GLP_MSG_ERR, GLP_MSG_ON, GLP_MSG_ALL
        glp_tm_lim = 10 * 1000
        glp_presolve = glpk.GLP_OFF       # or glpk.GLP_OFF (default) or GLP_ON
        self.res = {}

        self.nrows, self.ncols = constr_coefs_coo.shape
        assert col_bnds.shape == (2, self.ncols)
        assert row_bnds.shape == (self.nrows, 2)

        self.lp = glpk.glp_create_prob()

        # configure variables/columns and related bounds
        glpk.glp_add_cols(self.lp, self.ncols)
        for i, lb_ub in enumerate(col_bnds.T):
            var_type, lb, ub = self._get_variable_type(lb_ub)
            glpk.glp_set_col_bnds(self.lp, 1 + i, var_type, lb, ub)

        # configure rows/constraints
        glpk.glp_add_rows(self.lp, self.nrows)
        for i, lb_ub in enumerate(row_bnds):
            var_type, lb, ub = self._get_variable_type(lb_ub)
            glpk.glp_set_row_bnds(self.lp, 1 + i, var_type, lb, ub)
        n_coefs = len(constr_coefs_coo.row)
        ia = glpk.intArray(1 + n_coefs)
        ja = glpk.intArray(1 + n_coefs)
        ar = glpk.doubleArray(1 + n_coefs)
        for i in range(n_coefs):
            ia[1 + i] = int(1 + constr_coefs_coo.row[i])
            ja[1 + i] = int(1 + constr_coefs_coo.col[i])
            ar[1 + i] = constr_coefs_coo.data[i]
        glpk.glp_load_matrix(self.lp, n_coefs, ia, ja, ar)

        # configure options
        self._smcp = glpk.glp_smcp()
        glpk.glp_init_smcp(self._smcp)
        self._smcp.msg_lev = glp_msg_lev
        self._smcp.tm_lim = glp_tm_lim
        self._smcp.presolve = glp_presolve

    @staticmethod
    def _get_variable_type(bounds):
        """Determine variable type for upper/lower bounds.

        :param bounds: upper and lower bounds
        :type bounds: 1D ndarray of length 2
        :return var_type: glpk variable type to configure
        :rtype: integer constant
        :return lb: value of lower bound
        :rtype float
        :return ub: value of upper bound
        :rtype float
        """
        lb, ub = bounds
        if np.isfinite(lb) and np.isfinite(ub):
            if lb < ub:
                var_type = glpk.GLP_DB
            else:
                ub = 0.0
                var_type = glpk.GLP_FX
        elif not np.isfinite(lb) and not np.isfinite(ub):
            lb = 0.0
            ub = 0.0
            var_type = glpk.GLP_FR
        elif np.isfinite(ub):
            lb = 0.0
            var_type = glpk.GLP_UP
        else:
            ub = 0.0
            var_type = glpk.GLP_LO
        return var_type, lb, ub

    def set_obj_coefs(self, obj_coefs):
        """set coefficients for objective function

        configure coefficients that are not zero

        Note: glpk related indices start at 1 (not zero)
        :param obj_coefs: coefficients of objective function
        :type obj_coefs: 1D ndarray of shape (1, ncols)
        """
        assert len(obj_coefs) == self.ncols
        for i in range(self.ncols):
            if obj_coefs[i] != 0.0:
                glpk.glp_set_obj_coef(self.lp, int(1 + i), obj_coefs[i])

    def remove_obj_coefs(self, obj_coefs):
        """remove coefficients from objective function

        set coefficients that are not zero to zero in the liner problem

        Note: glpk related indices start at 1 (not zero)
        :param obj_coefs: coefficients of objective function
        :type obj_coefs: list-like of floats
        """
        assert len(obj_coefs) == self.ncols
        for i in range(self.ncols):
            if obj_coefs[i] != 0.0:
                glpk.glp_set_obj_coef(self.lp, int(1 + i), 0.0)

    def set_obj_dir(self, direction):
        """Configure the optimization direction

        :param direction: direction for the optimization ('maximize' or 'minimize')
        :type direction: string
        """
        obj_dir = glpk.GLP_MAX if 'max' in direction else glpk.GLP_MIN
        glpk.glp_set_obj_dir(self.lp, obj_dir)

    def add_constraint(self, constr_coefs, row_bnds):
        """add a single constraint to the lp problem

        :param constr_coefs: constraint coefficients for the new row
        :type constr_coefs: 1d ndarray of length ncols
        :param row_bnds: upper and lower bounds of the auxiliary (row) variable
        :type row_bnds: 1D ndarray of length 2
        :return: row index (start from 0)
        :rtype: int
        """
        row_idx1 = glpk.glp_add_rows(self.lp, 1)
        glpk.glp_set_row_bnds(self.lp, row_idx1, *self._get_variable_type(row_bnds))

        n_coef = len(np.nonzero(constr_coefs)[0])
        ind = glpk.intArray(n_coef + 1)
        val = glpk.doubleArray(n_coef + 1)
        for num_idx0, col_idx0 in enumerate(np.nonzero(constr_coefs)[0]):
            ind[num_idx0 + 1] = int(col_idx0 + 1)
            val[num_idx0 + 1] = constr_coefs[col_idx0]
        glpk.glp_set_mat_row(self.lp, row_idx1, n_coef, ind, val)
        self.nrows += 1
        row_idx0 = row_idx1 - 1
        return row_idx0

    def del_constraint(self, row_idx0):
        """remove single constraint from the lp problem

        :param row_idx0: row index (start from 0) to be removed
        :type row_idx0: int
        """
        assert row_idx0 < self.nrows
        num = glpk.intArray(1 + 1)
        num[1] = row_idx0 + 1
        glpk.glp_del_rows(self.lp, 1, num)
        self.nrows -= 1

    def set_col_bnd(self, col_idx0, bounds):
        """Set a single column (variable) bound
        :param col_idx0: column index (start at 0) of variable
        :type col_idx0: int
        :param bounds: lower and upper bound to set
        :type bounds: list with two float values
        """
        assert col_idx0 < self.ncols
        glpk.glp_set_col_bnds(self.lp, col_idx0 + 1, *self._get_variable_type(bounds))

    def remove_cols(self, col_idxs0):
        """ remove variables (columns) from the linear problem

        :param col_idxs0: col indices (start at 0) to be removed
        :type col_idxs0: list or 1D ndarray of col indices
        """
        num = glpk.intArray(len(col_idxs0) + 1)
        for num_idx0, col_idx0 in enumerate(col_idxs0):
            assert col_idx0 < self.ncols
            num[num_idx0 + 1] = col_idx0 + 1
        glpk.glp_del_cols(self.lp, len(col_idxs0), num)
        self.ncols -= len(col_idxs0)

    def remove_rows(self, row_idxs0):
        """ remove constraints (rows) from the linear problem.

        :param row_idxs0: row indices (start at 0) to be removed
        :type row_idxs0: list or 1D ndarray of row indices
        """
        num = glpk.intArray(len(row_idxs0) + 1)
        for num_idx0, row_idx0 in enumerate(row_idxs0):
            assert row_idx0 < self.nrows
            num[num_idx0 + 1] = row_idx0 + 1
        glpk.glp_del_rows(self.lp, len(row_idxs0), num)
        self.nrows -= len(row_idxs0)

    def construct_inital_basis(self):
        """Re-construct an initial basis, e.g. after problem update
        """
        glpk.glp_adv_basis(self.lp, 0)

    def solve(self, short_result=True, scaling=None):
        """Solve the Linear Programming Problem.

        :param short_result: short result (only objective value and status) or long result
        :type short_result: bool (optional, default: True)
        :param scaling: problem scaling options ('GM', 'EQ', '2N', 'SKIP', 'AUTO'
        :type scaling: string (optional, default: None)
        :return: result from linear optimization
        :rtype: dict
        """
        if scaling is not None:
            glpk.glp_scale_prob(self.lp, scaling_options.get(scaling, glpk.GLP_SF_AUTO))

        simplex_result = glpk.glp_simplex(self.lp, self._smcp)

        # in case of unsuccessful result, set basis and redo optimization
        if glpk.glp_get_status(self.lp) != glpk.GLP_OPT:
            glpk.glp_adv_basis(self.lp, 0)
            # print('GLP ADV BASIS')
            # glpk.glp_cpx_basis(self.lp)
            simplex_result = glpk.glp_simplex(self.lp, self._smcp)

        self.res = {
            'fun': glpk.glp_get_obj_val(self.lp),
            'success': glpk.glp_get_status(self.lp) == glpk.GLP_OPT,
            'generic_status': generic_status[glpk.glp_get_status(self.lp)],
        }
        if short_result is False:
            self.res['x'] = np.array([glpk.glp_get_col_prim(self.lp, 1 + i) for i in range(self.ncols)])
            self.res['reduced_costs'] = np.array([glpk.glp_get_col_dual(self.lp, 1 + i)
                                                  for i in range(self.ncols)])
            self.res['shadow_prices'] = np.array([glpk.glp_get_row_dual(self.lp, 1 + i)
                                                  for i in range(self.nrows)])
            self.res['simplex_result'] = simplex_result
            self.res['simplex_message'] = simplex_status[simplex_result],
            self.res['primal_status'] = prim_status[glpk.glp_get_prim_stat(self.lp)]
            self.res['dual_status'] = dual_status[glpk.glp_get_dual_stat(self.lp)]
        return self.res

    def get_row_value(self, row):
        """Retrieve right-hand side of a constraint (auxiliary variable)

        :param row: index into constraints
        :type row: int
        :return: value of auxiliary varibale (value of rhs)
        :rtype: float
        """
        assert row <= glpk.glp_get_num_rows(self.lp)
        return glpk.glp_get_row_prim(self.lp, row)

    def __del__(self):
        """Explicitly delete the LpProblem. (can we do without?)
        """
        # print('deletion of LpProblem')
        assert self.lp is not None
        if getattr(self, 'lp'):
            glpk.glp_delete_prob(self.lp)
